using CP;

// přirazení všech hodnot z datového souboru
int Dny = ...;
int Hodin = ...;
int Predmetu = ...;
int Trvani[1..Predmetu] = ...;
int Mistnosti = ...;
int Tridy = ...;

// pomocné range
range rTridy = 1..Tridy;
range rPredmety = 1..Predmetu;
range rMistnosti = 1..Mistnosti;

// tuple zaznamenávající přiřazení předmetu ke třídám
tuple predmetyTrid{
 int trida;
 int predmet;
}

// naplnění tuple z datového souboru
{predmetyTrid} PredmetyTrid = ...; 

// přiřazení času předmětům v maximálním rozmezí Dny*Hodin, přiřazení trvání
dvar interval casy[p in rPredmety] in 0 .. Dny * Hodin size Trvani[p];

// volitelný interval pro přiřazení předmětů a tříd do místnosti
dvar interval prirazeni[rMistnosti][PredmetyTrid] optional;

// sekvence rozvrhu tříd, z důvodu nepřekrývání
dvar sequence rozvrhTridy[t in rTridy] in all (m in rMistnosti, pt in PredmetyTrid : pt.trida == t) prirazeni[m][pt];

// sekvence rozvrhu místnosti, z důvodu nepřekrývání
dvar sequence rozvrhMistnosti[m in rMistnosti] in all (pt in PredmetyTrid) prirazeni[m][pt];

// minimalizace konce posledního předmětu přes všechny třídy
minimize max(p in rPredmety) endOf(casy[p]);

subject to {

// promazání domén, tak aby předmět nezačínal jeden den a končil druhý
forall(p in rPredmety)
  startOf(casy[p]) % Hodin <= (Hodin-Trvani[p]);

// zamezení výuky dvou předmětů pro jednu třídu zároveň
forall(t in rTridy)
  noOverlap(rozvrhTridy[t]);

// zamezení výuky dvou předmětů v jedné místnosti
forall(m in rMistnosti)
  noOverlap(rozvrhMistnosti[m]);

// pro každý předmět přiřazení jednoho volitelného intervalu (místnosti)
forall(pt in PredmetyTrid)
  alternative(casy[pt.predmet], all(m in rMistnosti)prirazeni[m][pt]);

}

// výpis pro kontrolu řešení
execute{
 for(var t=1; t <= Tridy; t++){
   writeln("Rozvrh "+t+". třídy:");
   for(var p in PredmetyTrid){
     if(p.trida==t){
       write("Předmět " + p.predmet + ": " + casy[p.predmet].start + "-" + casy[p.predmet].end);   
       for(var m=1; m<=Mistnosti; m++){
         if(prirazeni[m][p].present)writeln(" v " + m + ". místnosti");   
       }       
     }
   }      
   writeln();
 }
}
