
% =============================================================================
% predmety.pl
% =============================================================================

% spusteni
% :- main.

% ----------------------------------------------------------------------------

:- use_module(library(clpfd)).
:- use_module(library(lists)).

% ----------------------------------------------------------------------------

pocet_dnu(3).

pocet_hodin(10).

pocet_trid(3).

% p(Trida, Nazev, Trvani)
predmety( [ p(1,1,5),p(1,2,5),p(1,3,5),p(1,4,4),
            p(2,5,5),p(2,6,5),p(2,7,5),p(2,8,5),p(2,9,4),
            p(3,10,5),p(3,11,5),p(3,12,5),p(3,13,5),p(3,14,4) ]).

pocet_mistnosti(3).

% ----------------------------------------------------------------------------

main :- 
	% inicializace konstant
	predmety(Predmety),
	pocet_dnu(PocetDnu),
        pocet_hodin(PocetHodin),
	pocet_mistnosti(PocetMistnosti),
	PosledniSlot is PocetDnu*PocetHodin-1,

	% inicializace domenovych promennych
	length(Predmety,PocetPredmetu),
	length(Casy,PocetPredmetu),
	length(Mistnosti,PocetPredmetu),
        domain(Casy,0,PosledniSlot),
        domain(Mistnosti,1,PocetMistnosti),

	% vytvoreni datovych struktur pro cumulatives
        vytvor_machines(PocetMistnosti,0,Machines),
        vytvor_tasks(Predmety,Casy,Mistnosti,Tasks,Konce),

	% predmety nesmi zacinat jeden den a koncit druhy den
	denni_zlom(Tasks,PocetDnu,PocetHodin),

	% omezeni zajistujici, aby byl v kazde mistnosti v kazdem case nejvyse jeden predmet
        cumulatives(Tasks,Machines,[bound(upper),task_intervals(true)]),

	% predmety jedne tridy se nesmi prekryvat
	Predmety = [p(_,PrvniTrida,_)|_],
	predmety_tridy(Predmety,Tasks,PrvniTrida,[]),

	% vypocet ucelove funkce
	Funkce #>= 0, Funkce #=< PocetMistnosti*(PosledniSlot+1), % domena musi byt ohranicena
	% v ucelove funkci navic pro zvyseni efektivity vypoctu (a nalezeni reseni pro dane zadani)
        % nastavujeme dolni mez, ktera se spocita jako soucet dob trvani vsech trid
	ucelova_funkce(Konce,Predmety,PrvniTrida,0,_FunkcePrvniTridy,Funkce),

	% prohledavani
        append(Casy,Mistnosti,CasoProstor),
	last(CasoProstor, Funkce, Promenne),
	% pro labeling pouzijeme defaultni vyber hodnoty (vyber minimalni hodnoty z domeny), 
        % protoze minimalizuje umisteni predmetu do casu
        labeling([minimize(Funkce)],Promenne),
	
	% tisk
	print('Predmet(Trida): Mistnost x Od-Do'),
	my_print(Predmety,Casy,Mistnosti).

% ----------------------------------------------------------------------------

vytvor_tasks([],[],[],[],[]).
vytvor_tasks([p(_,_,Trvani)|Predmety],[C|Casy],[M|Mistnosti],
	     [task(C,Trvani,Konec,1,M)|Tasks],[Konec|Konce]) :-
	Konec #= C+Trvani,
        vytvor_tasks(Predmety,Casy,Mistnosti,Tasks,Konce).

vytvor_machines(PocetMistnosti,PocetMistnosti,[]) :- !.
vytvor_machines(PocetMistnosti,Minula,[machine(Aktualni,1)|Machines]) :-
	Aktualni is Minula+1,
        vytvor_machines(PocetMistnosti,Aktualni,Machines).

% ----------------------------------------------------------------------------

denni_zlom([],_,_).
denni_zlom([task(Cas,Trvani,_,_,_)|Tasks],PocetDnu,PocetHodin) :-
	cely_tyden(Cas,Trvani,0,PocetDnu,PocetHodin),
	denni_zlom(Tasks,PocetDnu,PocetHodin).

% zajistuje spravny denni zlom pro predmet s pocatecnim casem Cas
cely_tyden(_,_,PocetDnu,PocetDnu,_) :- !.
cely_tyden(Cas,Trvani,Dnes,PocetDnu,PocetHodin) :-
	Zitra is Dnes+1,
	kazdy_den(Zitra,Cas,PocetHodin,Trvani),
	cely_tyden(Cas,Trvani,Zitra,PocetDnu,PocetHodin).

% zajistuje spravny denni zlom pro predmet s pocatecnim casem Cas
% a pro den vyuky urceny promennou Zitra
kazdy_den(_,_,_,1) :-!.
kazdy_den(Zitra,Cas,PocetHodin,Rozdil) :-
	Cas #\= PocetHodin*Zitra - Rozdil + 1,
	NovyRozdil is Rozdil - 1, 
	kazdy_den(Zitra,Cas,PocetHodin,NovyRozdil).

% ----------------------------------------------------------------------------

% posilame cumulative na predmety posledni tridy
predmety_tridy([],[],_,AktualniTasks) :-
	cumulative(AktualniTasks).
% sbirame predmety jedne tridy, na ktere pak posleme cumulative
predmety_tridy([p(AktualniTrida,Nazev,_)|Predmety],[task(C,T,K,1,_)|Tasks],AktualniTrida,AktualniTasks) :-
	!,
	predmety_tridy(Predmety,Tasks,AktualniTrida,[task(C,T,K,1,Nazev)|AktualniTasks]).
% prosli jsme predmety jedne tridy, posilame cumulative
predmety_tridy([p(NovaTrida,Nazev,Trvani)|Predmety],Tasks,_,AktualniTasks) :-
	cumulative(AktualniTasks,[global(true)]),
	predmety_tridy([p(NovaTrida,Nazev,Trvani)|Predmety],Tasks,NovaTrida,[]).

% ----------------------------------------------------------------------------

% do Funkce zahrneme koncovy casy vyuky posledniho predmetu posledni tridy
% tj. do posledniho argumentu dame hodnotu FunkcePosledniTridy
ucelova_funkce([],[],_,SoucetTrvani,FunkcePosledniTridy,FunkcePosledniTridy) :-
        % FunkcePosledniTridy musi byt rovna alespon souctu dob trvani vsech predmetu posledni tridy
	FunkcePosledniTridy #>= SoucetTrvani.
% omezujeme cas ukonceni aktualni tridy
ucelova_funkce([K|Konce],[p(AktualniTrida,_,Trvani)|Predmety],AktualniTrida,SoucetTrvani,FunkceTridy,Funkce) :-
	!,
	K #=< FunkceTridy,
	% secitame doby trvani predmetu v aktualni tride
	NovySoucet is Trvani+SoucetTrvani,
	ucelova_funkce(Konce,Predmety,AktualniTrida,NovySoucet,FunkceTridy,Funkce).
% dostali jsme se k predmetum dalsi tridy nazvane NovaTrida
% secitame koncove casy vyuky poslednich predmetu kazde tridy
ucelova_funkce(Konce,[p(NovaTrida,Nazev,Trvani)|Predmety],_,SoucetTrvani,FunkceTridy,Funkce) :-
	Funkce #= FunkceTridy + ZbytekTrid,
        % FunkceTridy musi byt rovna alespon souctu dob trvani vsech predmetu teto tridy
	FunkceTridy #>= SoucetTrvani, 
	ucelova_funkce(Konce,[p(NovaTrida,Nazev,Trvani)|Predmety],NovaTrida,0,_FunkceNoveTridy,ZbytekTrid).

% ----------------------------------------------------------------------------

my_print([],[],[]) :- nl.
my_print([p(Trida,Nazev,Trvani)|Predmety],[C|Casy],[M|Mistnosti]) :- 
	nl,
	(Nazev<10 -> print(' '); true),
	print(Nazev), 
	print('('),
	print(Trida), 
	print('): '), 
	print(M),
	print(' x '),
	print(C),
	print('-'),
	Konec is C+Trvani,
	print(Konec),
	my_print(Predmety,Casy,Mistnosti).

