% =============================================================================
% Vzorove reseni problemu pro planovani projektu
% =============================================================================

% ----------------------------------------------------------------------------
% spousti se pomoci -? project.
% ----------------------------------------------------------------------------

:- use_module(library(clpfd)).
:- use_module(library(lists)).

% ----------------------------------------------------------------------------
% zadani dat

% job(Id, Duration, Person)
% predpokladame, ze Person jsou cislovany vzestupne od 1
job(1,3,1).
job(2,2,1).
job(3,1,1).  
job(4,4,1).  
job(5,2,1).  
job(6,4,2).  
job(7,2,2).  
job(8,1,2).  
job(9,1,2).  
job(10,3,2).  
job(11,2,3).  
job(12,2,3).  
job(13,1,3).  
job(14,1,3).

% limit(Person,Limit)
% osoba Person smi pracovat nejvyse na Limit ukolech zaroven
limit(1,2).
limit(2,2).
limit(3,1).

% same(K,L) ... K zacina stejne jako L
same(1,2).
same(7,8).

% prec(I,J) ... I zacina az po J
prec(3,1).
prec(6,7).

% ----------------------------------------------------------------------------
% hlavni predikat

project :-
        find_tasks(Tasks),
	make_lists(Tasks, Starts, Durations, Ends),
	sumlist(Durations, Max),
	domain(Starts, 0, Max),
	find_machines(Machines),
	cumulatives(Tasks,Machines,[bound(upper)]),

	precedence(Tasks),
	same_start(Tasks),

	sum(Ends, #=, F),
	append(Starts,[F],Vars),

        labeling([minimize(F),ffc],Vars),
	print('Optimum: '), print(F),nl,
	print_tasks(Tasks).

% ----------------------------------------------------------------------------
% pomocne predikaty

% vytvoreni seznamu Tasks
find_tasks(Tasks) :-
	findall( task(_S,Duration,_E,1,PersonId), 
                 job(_TaskId,Duration,PersonId),
                 Tasks ).

% vytvoreni seznamu Starts, Durations, Ends
make_lists([], [], [], []).
make_lists([task(S,D,E,1,_P)|Tasks], [S|Starts], [D|Durations], [E|Ends]):-
	E #= S+D,
	make_lists(Tasks, Starts, Durations, Ends).

% vytvoreni seznamu Machines
find_machines(Machines) :-
	findall( machine( Id, Limit ), limit( Id, Limit ), Machines ).

% omezujici podminky pro precedence
precedence(Tasks) :-
	findall(prec(First,Sec),prec(First,Sec),Prec),
	precedence(Prec, Tasks).

precedence([],_).
precedence([prec(First,Sec)|Prec], Tasks) :-
	nth1(First, Tasks, task(S,_,_,_,_)),
	nth1(Sec, Tasks, task(_,_,E,_,_)),
	S #>= E,
	precedence(Prec, Tasks).
	
% omezujici podminky pro stejny startovni cas
same_start(Tasks) :-
        findall(same(First,Sec),same(First,Sec),Same),
        same_start(Same, Tasks).

same_start([],_).
same_start([same(First,Sec)|Same], Tasks) :-
	nth1(First, Tasks, task(S1,_,_,_,_)),
	nth1(Sec, Tasks, task(S2,_,_,_,_)),
        S1 #= S2,
        same_start(Same, Tasks).

% ----------------------------------------------------------------------------
% pomocne vypisy

print_tasks(Tasks) :-
	make_keys(Tasks,1,KeyTasks),
	keysort(KeyTasks,PrintTasks),
	print('PersonId, TaskId: Start -> End\n'),
	print_tasks_out(PrintTasks).

make_keys([],_, []).
make_keys([task(S,D,E,1,P)|Tasks], I, [key(P,S)-task(S,D,E,I,P)|KeyTasks]) :-
	NewI is I + 1,
	make_keys(Tasks, NewI, KeyTasks).

print_tasks_out([]).
print_tasks_out([key(P,S)-task(S,_D,E,I,P)|Tasks]) :-
	format('       ~d, ~d: ~d -> ~d~n',[P,I,S,E]),
	print_tasks_out(Tasks).

% ----------------------------------------------------------------------------

end_of_file.

Vystup:

Optimum: 50
PersonId, TaskId: Start -> End
       1, 1: 0 -> 3
       1, 2: 0 -> 2
       1, 4: 2 -> 6
       1, 3: 3 -> 4
       1, 5: 4 -> 6
       2, 7: 0 -> 2
       2, 8: 0 -> 1
       2, 9: 1 -> 2
       2, 6: 2 -> 6
       2, 10: 2 -> 5
       3, 13: 0 -> 1
       3, 14: 1 -> 2
       3, 11: 2 -> 4
       3, 12: 4 -> 6
